#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <libporty/host/types.h>
#include "sockev.h"
#include "filters.h"

/* configuration is an integer (max buffer length) */

typedef struct packet_s {
	char *buff;
	P_int32_t plen; /* packet length (header) */
	P_int32_t got;  /* bytes received so far */
	P_int32_t maxlen;
} packet_t;

void sockev_filter_packet(sockev_t id, void **sockdata, sockev_f_dir_t event, void *data_, int len, int cookie)
{
	packet_t *p;
	char *data = data_;


	switch(event) {
		case SOCKEV_F_INIT:
			p = malloc(sizeof(packet_t));
			p->maxlen = *(P_int32_t *)data;
			p->buff = NULL;
			p->got = 0;
			p->plen = -1;
			*sockdata = p;
			break;

		case SOCKEV_F_IN:
			{
				p = *sockdata;
				if (p->buff == NULL)
					p->buff = malloc(p->maxlen);
				while (len > 0) {
					P_int32_t fits;
					if (p->plen < 0) {
						if ((len < 4) || (p->got != 0)) {
							/* partial receiving of len */
							while((p->got < 4) &&  (len > 0)) {
								p->buff[p->got] = *data;
								p->got++;
								data++;
								len--;
							}
							if (p->got < 4)
								return; /* not enough data to determine packet length */
							p->plen = ntohl(*(P_int32_t *)p->buff);
						}
						else {
							/* we didn't have partial length and our current read is at least 4 bytes - read length in one go */
							p->plen = ntohl(*(P_int32_t *)data);
							len -= 4;
							data += 4;
						}

						if (p->plen > p->maxlen) {
#warning TODO report error
							sockev_close(id);
							return;
						}
						p->got = 0;
					}
					if (len == 0)
						return;

					/* by now we have a length and maybe even a partial package in
					   p->buff - calculate how much more should be copied from current data */
					if (len <= (p->plen - p->got))
						fits = len;
					else
						fits = p->plen - p->got;

					/* copy and adjust data buffer and lengths */
					memcpy(p->buff + p->got, data, fits);
					p->got += fits;
					len -= fits;
					data += fits;

					/* check if a full packet is already accumulated in p */
					assert(p->got <= p->plen);
					if (p->got == p->plen) {
						sockev_filter_next(id, event, p->buff, p->plen, cookie);
						p->got = 0;
						p->plen = -1;
					}
				}
			}
			break;

		case SOCKEV_F_OUT:
			{
				P_int32_t l;

				p = *sockdata;
				if (len > p->maxlen) {
#warning TODO report error
					sockev_close(id);
					return;
				}

				/* send header (len) and data */
				l = htonl(len);
				sockev_filter_next(id, event, &l, 4, cookie);
				sockev_filter_next(id, event, data, len, cookie);
				break;
			}

		case SOCKEV_F_CLOSE:
			/* free buffer */
			p = *sockdata;
			if (p == NULL)
				break;
			if (p->buff != NULL)
				free(p->buff);
			free(p);
			*sockdata = NULL; /* paranoia... */
			break;

		case SOCKEV_F_OPEN:
			break;
	}
}
